==> cds-services-impl-2.7.0-sources.jar: com/sap/cds/services/impl/messages/MessageTargetImpl.java <==
/**************************************************************************
 * (C) 2019-2024 SAP SE or an SAP affiliate company. All rights reserved. *
 **************************************************************************/
package com.sap.cds.services.impl.messages;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Stack;
import java.util.function.Function;

import com.sap.cds.impl.builder.model.StructuredTypeProxy;
import com.sap.cds.impl.parser.PathParser;
import com.sap.cds.ql.CQL;
import com.sap.cds.ql.ElementRef;
import com.sap.cds.ql.StructuredType;
import com.sap.cds.ql.cqn.CqnReference;
import com.sap.cds.ql.cqn.CqnReference.Segment;
import com.sap.cds.ql.cqn.CqnStructuredTypeRef;
import com.sap.cds.ql.cqn.CqnVisitor;
import com.sap.cds.ql.cqn.Path;
import com.sap.cds.ql.cqn.ResolvedSegment;
import com.sap.cds.reflect.CdsElement;
import com.sap.cds.services.messages.MessageTarget;

public class MessageTargetImpl implements MessageTarget {

	private final String parameter;
	private final CqnReference reference;

	private MessageTargetImpl(String parameter, CqnReference reference) {
		this.parameter = parameter;
		this.reference = reference;
	}

	public static MessageTarget create(Path path, CdsElement element) {
		List<Segment> segments = new ArrayList<>();

		Iterator<ResolvedSegment> iterator = path.iterator();
		if (iterator.hasNext()) {
			iterator.next(); // skip root entity
		}
		iterator.forEachRemaining(resolvedSegment -> segments.add(resolvedSegment.segment()));
		segments.addAll(PathParser.segments(element.getName()));
		CqnReference ref;
		if (element.getType().isAssociation()) {
			ref = CQL.to(segments).asRef();
		} else {
			ref = CQL.get(segments).asRef();
		}
		return new MessageTargetImpl(MessageTarget.PARAMETER_CQN, ref);
	}

	public static MessageTarget create(String target) {
		return new MessageTargetImpl(target, null);
	}

	public static MessageTarget create(String parameter, Function<StructuredType<?>, Object> path) {
		return create(parameter, "SKIP", path);
	}

	public static MessageTarget create(String parameter, String rootElement, Function<StructuredType<?>, Object> path) {
		return create(parameter, CQL.to(rootElement), path);
	}

	@SuppressWarnings("unchecked")
	public static <E extends StructuredType<E>> MessageTarget create(String parameter, Class<E> type, Function<E, Object> path) {
		return create(parameter, StructuredTypeProxy.create(type), (Function<StructuredType<?>, Object>) path);
	}

	private static MessageTarget create(String parameter, StructuredType<?> rootElement, Function<StructuredType<?>, Object> path) {
		Object p = path.apply(rootElement);
		List<Segment> segments = new ArrayList<>();

		boolean structured = false;

		if(p instanceof StructuredType) {
			StructuredType<?> type = (StructuredType<?>) p;
			segments.addAll(type.get("SKIP").segments());
			segments.remove(segments.size() - 1);
			structured = true;
		} else if(p instanceof ElementRef) {
			segments.addAll(((ElementRef<?>) p).segments());
		} else {
			// invalid specification
			return null;
		}

		Stack<String> entityName = new Stack<>();
		rootElement.accept(new CqnVisitor() {
			@Override
			public void visit(CqnStructuredTypeRef typeRef) {
				entityName.push(typeRef.firstSegment());
			}
		});

		String name = entityName.pop();
		if(name.equals(segments.get(0).id())) {
			segments.remove(0);
		}

		if (structured) {
			return new MessageTargetImpl(parameter, CQL.to(segments).asRef());
		} else {
			return new MessageTargetImpl(parameter, CQL.get(segments).asRef());
		}
	}

	@Override
	public CqnReference getRef() {
		return reference;
	}

	@Override
	public String getParameter() {
		return parameter;
	}
}

==> cds-services-impl-2.10.6-sources.jar: com/sap/cds/services/impl/messages/MessageTargetImpl.java <==
/**************************************************************************
 * (C) 2019-2024 SAP SE or an SAP affiliate company. All rights reserved. *
 **************************************************************************/
package com.sap.cds.services.impl.messages;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Stack;
import java.util.function.Function;

import com.sap.cds.impl.builder.model.StructuredTypeProxy;
import com.sap.cds.impl.parser.PathParser;
import com.sap.cds.ql.CQL;
import com.sap.cds.ql.ElementRef;
import com.sap.cds.ql.StructuredType;
import com.sap.cds.ql.cqn.CqnReference;
import com.sap.cds.ql.cqn.CqnReference.Segment;
import com.sap.cds.ql.cqn.CqnStructuredTypeRef;
import com.sap.cds.ql.cqn.CqnVisitor;
import com.sap.cds.ql.cqn.Path;
import com.sap.cds.ql.cqn.ResolvedSegment;
import com.sap.cds.reflect.CdsElement;
import com.sap.cds.services.messages.MessageTarget;

public class MessageTargetImpl implements MessageTarget {

	private final String parameter;
	private final CqnReference reference;

	private MessageTargetImpl(String parameter, CqnReference reference) {
		this.parameter = parameter;
		this.reference = reference;
	}

	public static MessageTarget create(Path path, CdsElement element) {
		List<Segment> segments = new ArrayList<>();

		Iterator<ResolvedSegment> iterator = path.iterator();
		if (iterator.hasNext()) {
			iterator.next(); // skip root entity
		}
		iterator.forEachRemaining(resolvedSegment -> segments.add(resolvedSegment.segment()));
		segments.addAll(PathParser.segments(element.getName()));
		CqnReference ref;
		if (element.getType().isAssociation()) {
			ref = CQL.to(segments).asRef();
		} else {
			ref = CQL.get(segments).asRef();
		}
		return new MessageTargetImpl(MessageTarget.PARAMETER_CQN, ref);
	}

	public static MessageTarget create(String target) {
		return new MessageTargetImpl(target, null);
	}

	public static MessageTarget create(String parameter, Function<StructuredType<?>, Object> path) {
		return create(parameter, "SKIP", path);
	}

	public static MessageTarget create(String parameter, String rootElement, Function<StructuredType<?>, Object> path) {
		return create(parameter, CQL.to(rootElement), path);
	}

	@SuppressWarnings("unchecked")
	public static <E extends StructuredType<E>> MessageTarget create(String parameter, Class<E> type, Function<E, Object> path) {
		return create(parameter, StructuredTypeProxy.create(type), (Function<StructuredType<?>, Object>) path);
	}

	private static MessageTarget create(String parameter, StructuredType<?> rootElement, Function<StructuredType<?>, Object> path) {
		Object p = path.apply(rootElement);
		List<Segment> segments = new ArrayList<>();

		boolean structured = false;

		if(p instanceof StructuredType type) {
			segments.addAll(type.get("SKIP").segments());
			segments.remove(segments.size() - 1);
			structured = true;
		} else if(p instanceof ElementRef ref) {
			segments.addAll(ref.segments());
		} else {
			// invalid specification
			return null;
		}

		Stack<String> entityName = new Stack<>();
		rootElement.accept(new CqnVisitor() {
			@Override
			public void visit(CqnStructuredTypeRef typeRef) {
				entityName.push(typeRef.firstSegment());
			}
		});

		String name = entityName.pop();
		if(name.equals(segments.get(0).id())) {
			segments.remove(0);
		}

		if (structured) {
			return new MessageTargetImpl(parameter, CQL.to(segments).asRef());
		} else {
			return new MessageTargetImpl(parameter, CQL.get(segments).asRef());
		}
	}

	@Override
	public CqnReference getRef() {
		return reference;
	}

	@Override
	public String getParameter() {
		return parameter;
	}
}
